//------------------------------------------------------------------------------
//TTexture class
//Written by Ananth B. AKA HellSpawn
//------------------------------------------------------------------------------
unit GLTexture;

interface

uses
    Windows,
    Classes,
    SysUtils,
    JPEG,
    Graphics,
    OpenGL,
    GL,
    GLU,
    BaseTypes,
    //This file contains the OpenGL extension definitions
    GLExt;

const
     MINFILTER:GLenum=GL_LINEAR;
     MAGFILTER:GLenum=GL_LINEAR;

type
    //This will be the structure that will hold the basic image data that
    //we will give to OpenGL
    TTextureImage=record
                        ImageData:array of GLubyte;
                        BPP:GLuint;
                        Width,Height:GLuint;
                        TexID:GLuint;
    end;
    //This is the header for the uncompressed TGA that we will load
    TTGAHeader=array[0..11] of byte;
    //This is the header for the Quake2 WAL format.
    TWALHeader=record
                     Name:array[0..31] of char;
                     Width,Height:longword;
                     Offset:array[0..3] of longint;
                     NextName:array[0..31] of char;
                     Flags,
                     Contents,
                     Value:longword;
    end;
const
     TGAHeader:TTGAHeader=(0,0,2,0,0,0,0,0,0,0,0,0);
     MultiTextureAvailable:boolean=true;
     WAL_Brightness:integer=0;
type
    TTexture=class
                  private
                  protected
                  fTextureData:TTextureImage;
                  fMIPMapping:boolean;
                  fName:string;
                  function GetID:GLuint;
                  public
                  function LoadTexture(aFilename:string):boolean;
                  function LoadTGA(aFileName:string):boolean;
                  function LoadBitmap(aFilename:string):boolean;
                  function LoadWAL(aFilename,PaletteFilename:string):boolean;
                  function LoadJPEG(aFilename:string):Boolean;
                  procedure Enable;virtual;
                  procedure Disable;virtual;
                  procedure Transparency;
                  procedure Translucency;
                  procedure LightMap;
                  procedure Opaque;
                  destructor Destroy;override;
                  property ID:GLuint
                           read GetID;
                  property  Width:GLuint
                            read fTextureData.Width;
                  property Height:GLuint
                           read fTextureData.Height;
                  property MIPMapping:boolean
                           read fMIPMapping
                           write fMIPMapping;
                  property Name:string
                           read fName
                           write fName;
    end;

    TTextureUpdateType=(tuPlasma,tuCustom);
    TTextureUpdateEvent=procedure (var aTextureData:TTextureImage) of object;

    TPixel=record
                 R, G, B : Byte;
     end;
    TProceduralTexture=class(TTexture)
                            private
                            fTextureUpdate:TTextureUpdateType;
                            fOnTextureUpdate:TTextureUpdateEvent;
                            fAnimated:boolean;
                            fFrameInterval,
                            fLastTickCount:cardinal;

                            //variables used bu plasma update
                            Sr,Sc:array[0..7] of integer;
                            sinHalfTable,sinOneTable,sinTwoTable,sinFourTable:array[0..511] of integer;
                            procedure SetTextureUpdateType(const Value: TTextureUpdateType);
                            protected
                            procedure PlasmaTextureUpdate(var aTextureData:TTextureImage);
                            public
                            procedure Update(TickCount:cardinal);virtual;
                            procedure Initialize(aWidth,aHeight:integer);
                            property TextureUpdate:TTextureUpdateType
                                     read fTextureUpdate
                                     write SetTextureUpdateType;
                            property OnTextureUpdate:TTextureUpdateEvent
                                     read fOnTextureUpdate
                                     write fOnTextureUpdate;
                            property FrameInterval:cardinal
                                     read fFrameInterval
                                     write fFrameInterval;
                            property Animated:boolean
                                     read fAnimated
                                     write fAnimated;
    end;

    TAnimatedTexture=class(TTexture)
                          private
                          fTextures:array of TTexture;
                          fNumTextures:integer;
                          fFrameInterval,
                          fLastTickCount:cardinal;
                          fAnimated:boolean;
                          fCurrentTexture:integer;
                          protected
                          function GetTexture(i:integer):TTexture;
                          procedure SetTexture(i:integer;tex:TTexture);
                          procedure SetNumTextures(i:integer);
                          public
                          constructor Create;overload;
                          constructor Create(TexCount:integer);overload;
                          procedure Update(TickCount:cardinal);virtual;
                          procedure Enable;override;
                          procedure Disable;override;
                          property Animated:boolean
                                   read fAnimated
                                   write fAnimated;
                          property Textures[i:integer]:TTexture
                                   read GetTexture
                                   write SetTexture;default;
                          property NumTextures:integer
                                   read fNumTextures
                                   write SetNumTextures;
                          property FrameInterval:cardinal
                                   read fFrameInterval
                                   write fFrameInterval;
                          property CurrentTexture:integer
                                   read fCurrentTexture
                                   write fCurrentTexture;
    end;

    function InitMultiTexturing:boolean;
    procedure SetActiveTMU(Target:GLEnum);
    procedure MultiTexCoOrd2f(target:GLEnum;s,t:GLFloat);
    procedure MultiTexCoOrd4f(target:GLEnum;s,t,r,q:GLFloat);
var
   glActiveTextureARB:PFNGLACTIVETEXTUREARBPROC;
   glMultiTexCoOrd2fARB:PFNGLMULTITEXCOORD2FARBPROC;
   glMultiTexCoOrd4fARB:PFNGLMULTITEXCOORD4FARBPROC;
implementation
function InitMultiTexturing:boolean;
begin
     glActiveTextureARB:=wglGetProcAddress('glActiveTextureARB');
     glMultiTexCoOrd2fARB:=wglGetProcaddress('glMultiTexCoord2fARB');
     glMultiTexCoOrd4fARB:=wglGetProcaddress('glMultiTexCoord4fARB');
     result:=((@glActiveTextureARB<>nil) and (@glMultiTexCoOrd2fARB<>nil) and (@glMultiTexCoOrd4fARB<>nil));
     MultiTextureAvailable:=result;

     glActiveTextureARB(GL_TEXTURE0_ARB);
     glEnable(GL_TEXTURE_2D);
end;
procedure SetActiveTMU(Target:GLEnum);
begin
     if not MultiTextureAvailable then exit;
     glActiveTextureARB(Target);
end;
procedure MultiTexCoOrd2f(target:GLEnum;s,t:GLFloat);
begin
     if not MultiTextureAvailable then exit;
     glMultiTexCoOrd2fARB(target,s,t);
end;
procedure MultiTexCoOrd4f(target:GLEnum;s,t,r,q:GLFloat);
begin
     if not MultiTextureAvailable then exit;
     glMultiTexCoOrd4fARB(target,s,t,r,q);
end;

constructor TAnimatedTexture.Create;
begin
     inherited;
     fCurrentTexture:=-1;
     fAnimated:=true;
     fFrameInterval:=200;
     fLastTickCount:=0;
end;
constructor TAnimatedTexture.Create(TexCount:integer);
var
   i:integer;
begin
     inherited Create;
     fCurrentTexture:=-1;
     fAnimated:=true;
     fFrameInterval:=200;
     fLastTickCount:=0;
     SetNumTextures(TexCount);
     for i:=0 to TexCount-1 do
         fTextures[i]:=TTexture.Create;
end;
function TAnimatedTexture.GetTexture(i:integer):TTexture;
begin
     result:=fTextures[i];
end;
procedure TAnimatedTexture.SetTexture(i:integer;tex:TTexture);
begin
     fTextures[i]:=tex;
end;
procedure TAnimatedTexture.SetNumTextures(i:integer);
begin
     fNumTextures:=i;
     SetLength(fTextures,i);
end;
procedure TAnimatedTexture.Update(TickCount:cardinal);
begin
     if not fAnimated then exit;
     if fCurrentTexture=-1 then exit;
     if (TickCount-fLastTickCount)>=fFrameInterval then
     begin
          if (fCurrentTexture+1)<=(fNumTextures-1) then
             Inc(fCurrentTexture)
          else
              fCurrentTexture:=0;
          fLastTickCount:=TickCount;
     end;
end;
procedure TAnimatedTexture.Enable;
begin
     if fCurrentTexture=-1 then exit;
     fTextures[fCurrentTexture].Enable;
end;
procedure TAnimatedTexture.Disable;
begin
     if fCurrentTexture=-1 then exit;
     fTextures[fCurrentTexture].Disable;
end;

function TTexture.GetID:GLuint;
begin
     result:=fTextureData.TexID;
end;
procedure TTexture.Enable;
begin
     glBindTexture(GL_TEXTURE_2D,fTextureData.TexID);
end;
procedure TTexture.Disable;
begin
     glBindTexture(GL_TEXTURE_2D,0);
end;
procedure TTexture.Transparency;
begin
     glEnable(GL_BLEND);
     glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
end;
procedure TTexture.Translucency;
begin
     glEnable(GL_BLEND);
     glBlendFunc(GL_SRC_ALPHA, GL_ONE);
end;
procedure TTexture.LightMap;
begin
     glEnable(GL_BLEND);
     glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR);
end;
procedure TTexture.Opaque;
begin
     glDisable(GL_BLEND);
end;
destructor TTexture.Destroy;
begin
     glDeleteTextures(1,@fTextureData.TexID);
     inherited;
end;

function TTexture.LoadTGA(aFileName:string):boolean;
var
   hdr:TTGAHeader;
   useful:array[0..5] of byte;
   BytesPerPixel,
   ImageSize,
   temp,
   imagetype:GLuint;
   f:file;
   i:GLuint;
begin
     result:=false;
     imagetype:=GL_RGBA;
     if not FileExists(aFileName) then exit;

     Assign(f,aFileName);
     try
        Reset(f,1);
        BlockRead(f,hdr,SizeOf(hdr));
        if not CompareMem(@hdr[0],@TGAHeader[0],SizeOf(hdr)) then exit;
        BlockRead(f,useful,SizeOf(useful));
     except
           exit;
     end;
     fTextureData.Width:=useful[1]*256 + useful[0];
     fTextureData.Height:=useful[3]*256 + useful[2];
     if (fTextureData.Width<=0)or
        (fTextureData.Height<=0) or
        ((useful[4]<>24) and (useful[4]<>32)) then
        exit;
     fTextureData.BPP:=useful[4];
     BytesPerPixel:=fTextureData.BPP div 8;
     ImageSize:=fTextureData.Width*fTextureData.Height*BytesPerPixel;
     try
        SetLength(fTextureData.ImageData,ImageSize);
     except
           exit;
     end;
     try
        BlockRead(f,fTextureData.ImageData[0],ImageSize)
     except
           fTextureData.ImageData:=nil;
           exit;
     end;
     i:=0;
     while i<ImageSize do
     begin
          temp:=fTextureData.Imagedata[i];
          fTextureData.Imagedata[i]:=fTextureData.Imagedata[i+2];
          fTextureData.Imagedata[i+2]:=temp;
          Inc(i,BytesPerPixel);
     end;
     Close(f);

     glGenTextures(1,@fTextureData.TexID);
     glBindTexture(GL_TEXTURE_2D, fTextureData.TexID);
     if not fMIPMapping then
     begin
          if fTextureData.BPP=24 then
             imagetype:=GL_RGB;
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
          glTexImage2D(GL_TEXTURE_2D,0,imagetype,fTextureData.Width,fTextureData.Height,0,imagetype,GL_UNSIGNED_BYTE,@fTextureData.ImageData[0]);
     end
     else
     begin
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
          if fTextureData.BPP=24 then
             imagetype:=GL_RGB;
          gluBuild2DMipmaps(GL_TEXTURE_2D, 3, fTextureData.Width, fTextureData.Height, imagetype ,GL_UNSIGNED_BYTE, @fTextureData.ImageData[0]);
     end;
     result:=true;
end;
function TTexture.LoadBitmap(aFilename:string):boolean;
var
   FileHeader: BITMAPFILEHEADER;
   InfoHeader: BITMAPINFOHEADER;
   Palette: array of RGBQUAD;
   BitmapFile: THandle;
   BitmapLength: LongWord;
   PaletteLength: LongWord;
   ReadBytes: LongWord;
   Front: ^Byte;
   Back: ^Byte;
   Temp: Byte;
   I : Integer;
   Width, Height : Integer;
   pData : Pointer;

begin
     result :=FALSE;
     BitmapFile := CreateFile(PChar(aFilename), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
     if (BitmapFile = INVALID_HANDLE_VALUE) then
        exit;

     // Get header information
     ReadFile(BitmapFile, FileHeader, SizeOf(FileHeader), ReadBytes, nil);
     ReadFile(BitmapFile, InfoHeader, SizeOf(InfoHeader), ReadBytes, nil);

     // Get palette
     PaletteLength := InfoHeader.biClrUsed;
     SetLength(Palette, PaletteLength);
     ReadFile(BitmapFile, Palette, PaletteLength, ReadBytes, nil);
     if (ReadBytes <> PaletteLength) then
        exit;

     Width  := InfoHeader.biWidth;
     Height := InfoHeader.biHeight;
     BitmapLength := InfoHeader.biSizeImage;
     if BitmapLength = 0 then
       BitmapLength := Width * Height * InfoHeader.biBitCount Div 8;

     // Get the actual pixel data
     GetMem(pData, BitmapLength);
     ReadFile(BitmapFile, pData^, BitmapLength, ReadBytes, nil);
     if (ReadBytes <> BitmapLength) then
        exit;

     CloseHandle(BitmapFile);

     // Bitmaps are stored BGR and not RGB, so swap the R and B bytes.
     for I :=0 to Width * Height - 1 do
     begin
          Front := Pointer(Integer(pData^) + I*3);
          Back := Pointer(Integer(pData^) + I*3 + 2);
          Temp := Front^;
          Front^ := Back^;
          Back^ := Temp;
     end;
     fTextureData.Width:=Width;
     fTextureData.Height:=Height;
     fTextureData.ImageData:=pData;
     glGenTextures(1,@fTextureData.TexID);
     glBindTexture(GL_TEXTURE_2D, fTextureData.TexID);
     if not fMIPMapping then
     begin
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
          glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,fTextureData.Width,fTextureData.Height,0,GL_RGB,GL_UNSIGNED_BYTE,@fTextureData.ImageData[0])
     end
     else
     begin
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
          gluBuild2DMipmaps(GL_TEXTURE_2D, 3, fTextureData.Width, fTextureData.Height, GL_RGB,GL_UNSIGNED_BYTE, @fTextureData.ImageData[0]);
     end;

     result :=TRUE;
end;

function TTexture.LoadWAL(aFilename,PaletteFilename:string):boolean;
var
   TexFile,PalFile:file;
   header:TWALHeader;
   tempImageData:array of GLubyte;
   i,CurIndex:integer;
   Palette:array[0..255,0..2] of byte;
begin
     result:=false;
     if  not FileExists(aFilename) then exit;
     if not FileExists(PaletteFilename) then exit;

     Assign(TexFile,aFilename);
     Reset(TexFile,1);

     Assign(PalFile,PaletteFilename);
     Reset(PalFile,1);

     BlockRead(TexFile,header,SizeOf(header));
     fTextureData.Width:=header.Width;
     fTextureData.Height:=header.Height;

     SetLength(tempImagedata,header.Width*header.Height);
     SetLength(fTextureData.ImageData,header.Width*header.Height*3);

     BlockRead(TexFile,tempImageData[0],header.Width*header.Height);
     BlockRead(PalFile,Palette,SizeOf(Palette));

     if WAL_Brightness<>0 then
        for CurIndex:=0 to 255  do
        begin
             if Palette[CurIndex,0]+WAL_Brightness<255 then
                Inc(Palette[CurIndex,0],WAL_Brightness);
             if Palette[CurIndex,1]+WAL_Brightness<255 then
                Inc(Palette[CurIndex,1],WAL_Brightness);
             if Palette[CurIndex,2]+WAL_Brightness<255 then
                Inc(Palette[CurIndex,2],WAL_Brightness);
        end;
     i:=0;
     for CurIndex:=0 to header.Width*header.Height do
     begin
          fTextureData.ImageData[i]:=Palette[tempImageData[CurIndex]][0];
          Inc(i);
          fTextureData.ImageData[i]:=Palette[tempImageData[CurIndex]][1];
          Inc(i);
          fTextureData.ImageData[i]:=Palette[tempImageData[CurIndex]][2];
          Inc(i);
     end;
     Close(TexFile);
     Close(PalFile);

     glGenTextures(1,@fTextureData.TexID);
     glBindTexture(GL_TEXTURE_2D, fTextureData.TexID);
     if not fMIPMapping then
     begin
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
          glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,fTextureData.Width,fTextureData.Height,0,GL_RGB,GL_UNSIGNED_BYTE,fTextureData.ImageData)
     end
     else
     begin
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
          glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
          gluBuild2DMipmaps(GL_TEXTURE_2D, 3, fTextureData.Width, fTextureData.Height, GL_RGB,GL_UNSIGNED_BYTE, fTextureData.ImageData);
     end;

     result:=true;
end;
function TTexture.LoadJPEG(aFilename: String):Boolean;
var
  Data : Array of LongWord;
  W, Width : Integer;
  H, Height : Integer;
  BMP : TBitmap;
  JPG : TJPEGImage;
  C : LongWord;
  Line : ^LongWord;
begin
  result :=FALSE;
  JPG:=TJPEGImage.Create;

  try
    JPG.LoadFromFile(aFilename);
  except
    MessageBox(0, PChar('Couldn''t load JPG - "'+ aFilename +'"'), PChar('BMP Unit'), MB_OK);
    Exit;
  end;

  // Create Bitmap
  BMP:=TBitmap.Create;
  BMP.pixelformat:=pf32bit;
  BMP.width:=JPG.width;
  BMP.height:=JPG.height;
  BMP.canvas.draw(0,0,JPG);        // Copy the JPEG onto the Bitmap

  Width :=BMP.Width;
  Height :=BMP.Height;
  SetLength(Data, Width*Height);

  For H:=0 to Height-1 do
  Begin
    Line :=BMP.scanline[Height-H-1];   // flip JPEG
    For W:=0 to Width-1 do
    Begin
      c:=Line^ and $FFFFFF; // Need to do a color swap
      Data[W+(H*Width)] :=(((c and $FF) shl 16)+(c shr 16)+(c and $FF00)) or $FF000000;  // 4 channel.
      inc(Line);
    End;
  End;

  fTextureData.Width:=Width;
  fTextureData.Height:=Height;
  fTextureData.ImageData:=addr(Data[0]);

  glGenTextures(1,@fTextureData.TexID);
  glBindTexture(GL_TEXTURE_2D, fTextureData.TexID);
  if not fMIPMapping then
  begin
       glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
       glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
       glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,fTextureData.Width,fTextureData.Height,0,GL_RGBA,GL_UNSIGNED_BYTE,fTextureData.ImageData)
  end
  else
  begin
       glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
       glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
       gluBuild2DMipmaps(GL_TEXTURE_2D, 3, fTextureData.Width, fTextureData.Height, GL_RGBA,GL_UNSIGNED_BYTE, fTextureData.ImageData);
  end;

  BMP.free;
  JPG.free;

  result :=TRUE;
end;

function TTexture.LoadTexture(aFilename: string): boolean;
var
   ext:string;
begin
     ext:=UpperCase(ExtractFileExt(aFilename));
     if ext='.JPG' then
     begin
          result:=LoadJPEG(aFilename);
          exit;
     end;
     if ext='.BMP' then
     begin
          result:=LoadBitmap(aFilename);
          exit;
     end;
     if ext='.TGA' then
     begin
          result:=LoadTGA(aFilename);
          exit;
     end;
     if ext='.WAL' then
     begin
          result:=LoadWAL(aFilename,ExtractFilepath(aFilename)+'palette.pal');
          exit;
     end
     else
         result:=false;
end;

{ TProceduralTexture }

procedure TProceduralTexture.Initialize(aWidth, aHeight: integer);
var
   i:integer;
   temp:single;
begin
     fTextureData.Width:=aWidth;
     fTextureData.Height:=aHeight;
     SetLength(fTextureData.ImageData,aWidth*aHeight*3);
     //GetMem(fTextureData.ImageData,aWidth*aHeight*3);
     //FillChar(fTextureData.Imagedata[0],aWidth*aHeight*3,128);
     glGenTextures(1,@fTextureData.TexID);
     glBindTexture(GL_TEXTURE_2D, fTextureData.TexID);
     glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,MINFILTER);
     glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,MAGFILTER);
     glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,fTextureData.Width,fTextureData.Height,0,GL_RGB,GL_UNSIGNED_BYTE,fTextureData.ImageData);

     Randomize;
     for i :=0 to 7 do
     begin
          sr[i] :=Random(256);
          sc[i] :=sr[i];
     end;

     for i :=0 to 511 do
     begin
          Temp :=4*pi*i/512;
          sinHalfTable[i] := Round(sin(temp/2)*128 + 128);
          sinOneTable[i]  := Round(sin(temp  )*128 + 128);
          sinTwoTable[i]  := Round(sin(temp*2)*128 + 128);
          sinFourTable[i] := Round(sin(temp*4)*128 + 128);
     end;
end;

procedure TProceduralTexture.PlasmaTextureUpdate(var aTextureData: TTextureImage);
var X, XCount, Y, YCount : Array[0..3] of Word;
    I, J, c : Integer;
    Pixel : ^TPixel;
begin
  for C :=0 to 3 do
    Y[c] :=sr[c+4];

  for C :=0 to 3 do
  begin
    XCount[c] :=sc[c];
    YCount[c] :=sc[c+4];
  end;

  for I :=0 to 63 do
  begin
    // reset the x values & counters
    for C :=0 to 3 do
      X[c] :=sr[c];

    for C :=0 to 3 do
    begin
      XCount[c] :=(sc[c] + I*2) MOD 512;
      YCount[c] :=(sc[c+4] + I*2) MOD 512;
    end;

    for J :=0 to 63 do
    begin
      // Set the Pixel Values
      Pixel :=Pointer(Integer(@fTextureData.ImageData[0]) + 3*(i*64 + j));
      Pixel.R :=Round((sinHalfTable[X[0]] + sinOneTable[X[1]] + sinTwoTable[X[2]] + sinFourTable[X[3]]) +
                      (sinHalfTable[Y[0]] + sinOneTable[Y[1]] + sinTwoTable[Y[2]] + sinFourTable[Y[3]])) SHR 3;
      Pixel.G :=Round((sinHalfTable[X[2]] + sinOneTable[X[3]] + sinTwoTable[X[0]] + sinFourTable[X[1]]) +
                      (sinHalfTable[Y[2]] + sinOneTable[Y[3]] + sinTwoTable[Y[0]] + sinFourTable[Y[1]])) SHR 3;
      Pixel.B :=Round((sinHalfTable[X[3]] + sinOneTable[X[0]] + sinTwoTable[X[1]] + sinFourTable[X[2]]) +
                      (sinHalfTable[Y[1]] + sinOneTable[Y[2]] + sinTwoTable[Y[3]] + sinFourTable[Y[0]])) SHR 3;

      // increment the x values
      for C :=0 to 3 do
      begin
        X[c] := Round((X[c] + (sinTwoTable[XCount[c]])/32 -4 + 512)) MOD 512;
        XCount[c] := Round(XCount[c] + 2) MOD 512;
      end;
    end;

    // increment the y values
    for C:=0 to 3 do
      Y[c] := Round(Y[c] + (sinOneTable[YCount[c]])/32 -4 + 512) MOD 512;
  end;

  // update statics
  for c :=0 to 7 do
  begin
    sr[c] := Round(sr[c] + (sinTwoTable[sc[c]])/32 -4 + 512) MOD 512;
    sc[c] := Round(sc[c] + 2) MOD 512;
  end;
  glTexImage2D(GL_TEXTURE_2D, 0, 3, 64, 64, 0, GL_RGB, GL_UNSIGNED_BYTE, @fTextureData.Imagedata[0]);
end;

procedure TProceduralTexture.SetTextureUpdateType(
  const Value: TTextureUpdateType);
begin
     fTextureUpdate:=Value;
     case fTextureUpdate of
          tuPlasma:fOnTextureUpdate:=PlasmaTextureUpdate;
          tuCustom:fOnTextureUpdate:=nil;
     end;
end;

procedure TProceduralTexture.Update(TickCount:cardinal);
begin
     if not fAnimated then exit;
     if (TickCount-fLastTickCount)>=fFrameInterval then
     begin
          if assigned(fOnTextureUpdate) then
             fOnTextureUpdate(fTextureData);
          fLastTickCount:=TickCount;
     end;
end;

end.
